import pytest
from aiomoqt.protocol import MOQTSessionProtocol
from aiomoqt.client import MOQTClientSession
from aiomoqt.types import *
from aiomoqt.messages import *


def test_moqt_stub():
    assert True